import re
import logging


class Rule:
    def __init__(self, name: str, funct: object) -> object:
        self.name = name
        self.funct = funct

    def run(self, x) -> bool:
        logging.debug("running rule {} on {}".format(self.name, x))
        return self.funct(x)


no_error = Rule("Check that dictionary json does not have error",
                lambda x: x and (type(x) != dict or "error" not in x))


type_str = Rule("Type is string", lambda x: x is not None and type(x) == str)
type_int = Rule("Type is integer", lambda x: x is not None and type(x) == int)
type_bool = Rule("Type is boolean", lambda x: x is not None and type(x) == bool)  # check not None first
type_dct = Rule("Type is dictionary", lambda x: x is not None and type(x) == dict)
type_lst = Rule("Type is list", lambda x: x is not None and type(x) == list)
type_float = Rule("Type is float", lambda x: x is not None and type(x) in [int, float])

type_url = Rule("Check url exists/ valid with https://www.",
                lambda x: x is not None and len(re.findall(r"^https://www\.(.+)\.(.+)", str(x))) > 0)
type_https = Rule("Check url exists/ valid with http", lambda x: x is not None and len(re.findall(r"^https://(.+)\.(.+)", str(x))) > 0)
type_domain = Rule("Check for exists/ valid domain", lambda x: x is not None and len(re.findall(r"(.+)\.(.+)", str(x))) > 0)
type_date = Rule("Check for exists/ valid date format", lambda x: x is not None and x)  # TODO: update this!
type_ticker = Rule("Check ticker exists/ consists of letters and possible dot",
                   lambda x: x is not None and re.findall(r"^\w+(\.)?\w+", str(x)))

non_empty_field = Rule("Check given field is non-empty", lambda x: x is not None and x)
not_none = Rule("Check field is not None", lambda x: x is not None)

frequency_name = Rule("Check valid frequency name",
                      lambda x: x is not None and x in ["instant", "daily", "weekly", "monthly", "quarterly", "yearly"])


def has_field(field_name):  # uses string name (field in x)
    return Rule("Has field `{}`".format(field_name), lambda x: x is not None and field_name in x)


def in_field(field_names):  # uses string name (x in in fields)
    return Rule("Within fields `{}`".format(field_names), lambda x: x is not None and x in field_names)


def len_eq(number):
    return Rule("Has field length `{}`".format(number), lambda x: x is not None and len(x) == number)


def len_gt(number):
    return Rule("Has field length greater than `{}`".format(number), lambda x: x is not None and len(x) > number)


def len_lt(number):
    return Rule("Has field length less than `{}`".format(number), lambda x: x is not None and len(x) < number)


def gte(number):
    return Rule("Is greater than or equal to `{}`".format(number), lambda x: x is not None and x >= number)


def lte(number):
    return Rule("Is less than or equal to `{}`".format(number), lambda x: x is not None and x <= number)


def matches_url(field_name):
    return Rule("Has url for `{}`".format(field_name), lambda x: x is not None and matches_url_funct(field_name, x))


def matches_url_funct(field, url):
    return len(re.findall(r"https://(www\.)?{name}\.com.*".format(name=field.name), str(url))) > 0



import datetime
import logging
from dateutil.relativedelta import relativedelta
from dateutil import parser
from .consts import daily_string, weekly_string, monthly_string, quarterly_string, \
    annually_string
from . import rule
import datetime
import pytz

utc=pytz.UTC


max_value = 1000 * 1000 * 1000 * 1000
max_delay = 1000 * 1000 * 1000 * 1000 * 1000 * 1000
value_default = None


class Field:
    def __init__(self):
        self.name = ""
        self.valid = False
        self.value = None
        self.rules = []
        self.errors = []
        self.fields = []

    def set(self, value):
        self.value = value
        return  # returns any errors

    def set_field(self, field, value):
        # for objects with field objects inside of them
        # bubbles up the errors
        field.set(value)
        return

    def set_fields(self, fields):
        # if using this then can run validate early bc field validation included
        # if not necessarily calling ste field on every field then must validate explicitly
        for field in fields:
            found_value = self.find_field(field)
            if found_value is not None:  # found_value may equal False
                self.set_field(field, found_value)
        return

    def get_field(self, name: str):
        for field in self.fields:
            if field.name == name:
                return field
        return

    def get_fields(self, names: list):
        fields = []
        for field in self.fields:
            if field.name in names:
                fields.append(field)
        return fields

    def find_field(self, field):
        if field.name in self.value:
            return self.value[field.name]
        else:
            self.error("field {} not found in object {}".format(field.name, self.name))
        return ""

    def run_rules(self):
        for rule in self.rules:
            ok = rule.run(self.value)
            if not ok:
                if type(self.value) in [str, int, bool]:
                    self.error("{}='{}' failed rule {}".format(self.name, self.value, rule.name))
                else:
                    self.error("{} with type {} failed rule {}".format(self.name, str(type(self.value)), rule.name))
        return

    def validate(self) -> list:
        """
        :param:
        :return: errors: list
        """
        self.run_rules()
        if self.errors:
            return self.errors
        self.validate_fields(self.fields)
        if not self.errors:
            self.valid = True
        return self.errors

    def validate_field(self, field) -> list:
        """
        :param: field: the actual field in the object
        :return: errors: list
        """
        error = field.validate()
        self.error_many(error)
        return error

    def validate_fields(self, fields):
        for field in fields:
            self.validate_field(field)
        return

    def error(self, msg: str):
        if msg not in self.errors:
            logging.error(msg)  # ERROR:root:error message
            self.errors.append(msg)
        return

    def error_many(self, msgs: list):
        if msgs:
            self.errors += msgs
        return


class PosIntField(Field):
    def __init__(self, name):
        super().__init__()
        self.name = name
        self.type = int
        self.value = value_default
        self.rules = [rule.not_none, rule.type_int, rule.gte(0), rule.lte(max_value)]


class IntField(Field):
    def __init__(self, name):
        super().__init__()
        self.name = name
        self.type = int
        self.value = value_default
        self.rules = [rule.not_none, rule.type_int, rule.gte(-max_value), rule.lte(max_value)]


class FloatField(Field):
    def __init__(self, name):
        super().__init__()
        self.name = name
        self.type = float
        self.value = value_default
        self.rules = [rule.not_none, rule.type_float, rule.gte(-max_value), rule.lte(max_value)]


class BigIntField(IntField):
    def __init__(self, name):
        super().__init__(name)
        self.rules = [rule.not_none, rule.type_int, rule.gte(0), rule.lte(max_delay)]

    def set(self, value):
        try:
            num = int(value)
            self.value = num
        except:
            self.error("cannot parse value {} as int".format(value))


class BoolField(Field):
    def __init__(self, name):
        super().__init__()
        self.name = name
        self.type = bool
        self.value = None
        self.rules = [rule.not_none, rule.type_bool]


class StringField(Field):
    def __init__(self, name):
        super().__init__()
        self.name = name
        self.type = str
        self.value = ""
        self.rules = [rule.not_none, rule.type_str, rule.len_gt(0), rule.len_lt(200)]


class EnumField(StringField):
    def __init__(self, name: str, names: list):
        super().__init__(name)
        self.rules = [rule.in_field(names)]


class URLField(StringField):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.rules = [rule.not_none, rule.type_str, rule.type_url]


class HttpsField(StringField):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.rules = [rule.not_none, rule.type_str, rule.type_https]


class DomainField(StringField):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.rules = [rule.not_none, rule.type_str, rule.type_domain]  # TODO: add normalizer rules


class TickerField(StringField):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.rules = [rule.not_none, rule.type_str, rule.type_ticker]


class DateField(Field):
    def __init__(self, name):
        super().__init__()
        self.name = name
        self.value = datetime.datetime(1000, 1, 1)
        self.rules = []

    def set(self, value):
        try:
            date = parser.parse(value)  # 2016-05-15T00:00:00Z
            # date = datetime.strptime(value, '%Y-%m-%dT%H:%M:%S.Z') # 2019-03-22T14:20:58.003Z
        except ValueError as err:
            self.error("Cannot parse date {}".format(value))
            return
        self.value = date
        self.run_rules()
        return self.errors

    def after(self, other_date):
        # self.value > other.value
        val1 = self.value.replace(tzinfo=utc)
        val2 = other_date.value.replace(tzinfo=utc)
        return val1 == max(val1, val2)

    def before(self, other_date):
        # self.value < other.value
        val1 = self.value.replace(tzinfo=utc)
        val2 = other_date.value.replace(tzinfo=utc)
        return val1 == min(val1, val2)

    def add_frequency(self, frequency: str):
        # Shallow Add. Returns a copy. Does not affect the original.
        copy = DateField(self.name)
        copy.rules = self.rules
        copy.value = self.value
        if frequency == daily_string:
            copy.value = self.value + relativedelta(days=+1)
        if frequency == weekly_string:
            copy.value = self.value + relativedelta(days=+7)
        if frequency == monthly_string:
            copy.value = self.value + relativedelta(months=+1)
        if frequency == quarterly_string:
            copy.value = self.value + relativedelta(months=+3)
        if frequency == annually_string:
            copy.value = self.value + relativedelta(years=+1)
        return copy




import re
from .field import Field, URLField, StringField, TickerField, HttpsField
from . import rule


def matches_url(field, url):
    return len(re.findall(r"https://(www\.)?{name}\.com.*".format(name=field.name), str(url))) > 0


def matches(field, sample):
    return len(re.findall(r".*{name}.*".format(name=field.name), str(sample))) > 0


class CAResponseField(Field):
    def __init__(self):
        super().__init__()
        self.name = "response"
        self.value = {}

        self.biids = BiidsField("biids")
        self.mainDomains = MainDomainField("mainDomains")
        self.urls = UrlsField("URLs")
        self.rules = [
            rule.no_error,
            rule.has_field(self.biids.name),
            rule.has_field(self.mainDomains.name),
            rule.has_field(self.urls.name)
        ]
        self.fields = [self.biids, self.mainDomains, self.urls]

    def set(self, value):
        self.value = value
        self.run_rules()  # checks that names exist in dictionary
        if self.errors:
            return self.errors
        self.set_fields(self.fields)  # goes through fields & sets them
        if self.errors:
            return self.errors
        return self.errors


class CAResponseSubField(Field):
    def __init__(self, name):
        super().__init__()
        self.name = name
        self.found = {}

    def was_found(self, field):
        if field.name in self.found:
            self.error("Duplicate {} url".format(field.name))
            return True
        self.found[field.name] = True
        return False


class BiidsField(CAResponseSubField):
    def __init__(self, name):
        super().__init__(name)
        self.datasys_id = StringField("datasys")
        self.timeseries_id = StringField("timeseries")
        self.ticker = TickerField("ticker")
        self.exchange = StringField("exchange")  # Todo: update to be more specific
        self.eod = TickerField("eod")
        self.rules = [
            # rule.no_error,
            rule.len_gt_zero
        ]
        self.fields = [self.datasys_id, self.timeseries_id, self.ticker, self.exchange, self.eod]

    def set(self, value):
        self.value = value
        self.run_rules()  # checks basics
        if self.errors:
            return self.errors
        for obj in self.value:
            value = obj["value"]
            service = obj["service"]
            if matches(self.datasys_id, service):
                if not self.was_found(self.datasys_id):
                    self.set_field(self.datasys_id, value)
            elif matches(self.timeseries_id, service):
                if not self.was_found(self.timeseries_id):
                    self.set_field(self.timeseries_id, value)
            elif matches(self.eod, service):
                if not self.was_found(self.eod):
                    self.set_field(self.eod, value)
            elif matches(self.exchange, service):  # Note: not limited to 1
                self.set_field(self.exchange, service)
                self.set_field(self.ticker, value)
        return self.errors


class MainDomainField(CAResponseSubField):
    def __init__(self, name):
        super().__init__(name)
        self.domain = StringField("mainDomain")  # TODO: is this right?
        self.rules = [
            # rule.no_error,
            rule.len_gt_zero
        ]
        self.fields = [self.domain]

    def set(self, value):
        self.value = value
        self.run_rules()  # checks basics
        if self.errors:  # returns any errors
            return self.errors
        for obj in self.value:
            if not self.was_found(self.domain):
                self.set_field(self.domain, obj["value"])
        return self.errors


class UrlsField(CAResponseSubField):
    def __init__(self, name):
        super().__init__(name)
        self.linkedin = URLField("linkedin")
        self.indeed = URLField("indeed")
        self.comparably = URLField("comparably")
        self.marketbeat = URLField("marketbeat")
        self.yahoo = HttpsField("finance.yahoo")
        self.twitter = HttpsField("twitter")
        self.facebook = URLField("facebook")
        self.rules = [
            # rule.no_error,
            rule.len_gt_zero
        ]
        self.fields = [self.linkedin, self.indeed, self.comparably, self.marketbeat, self.yahoo, self.twitter, self.facebook]

    def set(self, value):
        self.value = value
        self.run_rules()  # checks basics
        if self.errors:  # returns any errors
            return self.errors
        for obj in self.value:
            url = obj["value"]
            for field in self.fields:
                if matches_url(field, url):
                    if not self.was_found(field):
                        self.set_field(field, url)
                    break
            else:
                self.error("`URLs` field has url {} which doesn't match expected sources/formats".format(url))
        return self.errors





import logging
from . import rule
from .timeseries_metadata import TimeseriesMetadataField
from .data_list import DataList
from .field import Field


class TimeseriesCheck(Field):
    def __init__(self):
        """
        :param metadata_names:
        :param rules: list of Rule objects
        """
        super().__init__()
        self.name = "Timeseries Check"
        self.value = {}

        self.metadata = TimeseriesMetadataField()
        self.data = DataList()
        self.fields = [self.metadata, self.data]  # will add fields when needed for use case
        self.rules = [rule.has_field(x.name) for x in self.fields]

    def set(self, value):
        self.value = value
        self.run_rules()
        if self.errors:  # returns any errors
            return self.errors
        self.metadata.rules = []  # default
        self.set_fields(self.fields)
        logging.debug("TimeseriesCheck: data field is set as {}".format(self.data))
        return self.errors

    def run_check(self, new_check, metadata_names=[]):
        """
        Returns errors only from that check
        :param new_check:
        :param metadata_names:
        :return:
        """
        self.metadata.rules = [rule.has_field(name) for name in metadata_names]
        errors = self.validate_field(self.metadata)  # automatically adds new errors to self.errors
        if errors:  # new errors from check
            return
        error = new_check.run(self.data.value, self.metadata)
        if error:
            self.error("Failed check {}: {}".format(new_check.name, error))
        return





from .field import Field
from .ca import CAResponseField
from .check import Check, CheckEach
from .data_list import DataList
from .data_object import DataObject
from .date_check import DateCheck
from .rule import Rule
from .timeseries_check import TimeseriesCheck
from .timeseries_metadata import TimeseriesMetadataField
from .timeseries_response import TimeseriesResponseField
from . import consts




# metadata fields
id_string = "id"
calculated_string = "calculated"
frequency_string = "frequency"
min_value_string = "minValue"
max_value_string = "maxValue"
start_time_string = "startTime"
is_integer_string = "isInteger"
delay_string = "delay"

# frequencies
instant_string = "instant"
daily_string = "daily"
weekly_string = "weekly"
monthly_string = "monthly"
quarterly_string = "quarterly"
annually_string = "annually"

frequencies = [instant_string, daily_string, weekly_string, monthly_string, quarterly_string, annually_string]
